from enum import Enum

import numpy as np
from copy import deepcopy
from rlcard.games.limitholdem import Game
from rlcard.games.limitholdem import PlayerStatus

from rlcard.games.nolimitholdem import Dealer
from rlcard.games.nolimitholdem import Player
from rlcard.games.nolimitholdem import Judger
from rlcard.games.nolimitholdem import Round, Action


class Stage(Enum):

    PREFLOP = 0
    FLOP = 1
    TURN = 2
    RIVER = 3
    END_HIDDEN = 4
    SHOWDOWN = 5


class NolimitholdemGame(Game):

    def __init__(self, allow_step_back=False, num_players=2):
        ''' Initialize the class nolimitholdem Game
        '''
        self.allow_step_back = allow_step_back
        self.np_random = np.random.RandomState()

        # small blind and big blind
        self.small_blind = 1
        self.big_blind = 2 * self.small_blind

        # config players
        self.num_players = num_players
        self.init_chips = 100

    def init_game(self):
        ''' Initialilze the game of Limit Texas Hold'em

        This version supports two-player limit texas hold'em

        Returns:
            (tuple): Tuple containing:

                (dict): The first state of the game
                (int): Current player's id
        '''
        # Initilize a dealer that can deal cards
        self.dealer = Dealer(self.np_random)

        # Initilize two players to play the game
        self.players = [Player(i, self.init_chips, self.np_random) for i in range(self.num_players)]

        # Initialize a judger class which will decide who wins in the end
        self.judger = Judger(self.np_random)

        # Deal cards to each  player to prepare for the first round
        for i in range(2 * self.num_players):
            self.players[i % self.num_players].hand.append(self.dealer.deal_card())

        # Initilize public cards
        self.public_cards = []
        self.stage = Stage.PREFLOP

        # Randomly choose a big blind and a small blind
        s = self.np_random.randint(0, self.num_players)
        b = (s + 1) % self.num_players
        self.players[b].bet(chips=self.big_blind)
        self.players[s].bet(chips=self.small_blind)

        # The player next to the small blind plays the first
        self.game_pointer = (b + 1) % self.num_players

        # Initilize a bidding round, in the first round, the big blind and the small blind needs to
        # be passed to the round for processing.
        self.round = Round(self.num_players, self.big_blind, dealer=self.dealer, np_random=self.np_random)

        self.round.start_new_round(game_pointer=self.game_pointer, raised=[p.in_chips for p in self.players])

        # Count the round. There are 4 rounds in each game.
        self.round_counter = 0

        # Save the hisory for stepping back to the last state.
        self.history = []

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def get_legal_actions(self):
        ''' Return the legal actions for current player

        Returns:
            (list): A list of legal actions
        '''
        return self.round.get_nolimit_legal_actions(players=self.players)

    def step(self, action):
        ''' Get the next state

        Args:
            action (str): a specific action. (call, raise, fold, or check)

        Returns:
            (tuple): Tuple containing:

                (dict): next player's state
                (int): next plater's id
        '''

        if action not in self.get_legal_actions():
            print(action, self.get_legal_actions())
            print(self.get_state(self.game_pointer))
            raise Exception('Action not allowed')

        if self.allow_step_back:
            # First snapshot the current state
            r = deepcopy(self.round)
            b = self.game_pointer
            r_c = self.round_counter
            d = deepcopy(self.dealer)
            p = deepcopy(self.public_cards)
            ps = deepcopy(self.players)
            self.history.append((r, b, r_c, d, p, ps))

        # Then we proceed to the next round
        self.game_pointer = self.round.proceed_round(self.players, action)

        players_in_bypass = [1 if player.status in (PlayerStatus.FOLDED, PlayerStatus.ALLIN) else 0 for player in self.players]

        # If a round is over, we deal more public cards
        if self.round.is_over():
            # For the first round, we deal 3 cards
            if self.round_counter == 0:
                self.stage = Stage.FLOP
                self.public_cards.append(self.dealer.deal_card())
                self.public_cards.append(self.dealer.deal_card())
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1
                    self.stage = Stage.TURN
                    self.public_cards.append(self.dealer.deal_card())
                    self.round_counter += 1
                    self.stage = Stage.RIVER
                    self.public_cards.append(self.dealer.deal_card())
                    self.round_counter += 1
            # For the following rounds, we deal only 1 card
            elif self.round_counter == 1:
                self.stage = Stage.TURN
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1
                    self.stage = Stage.RIVER
                    self.public_cards.append(self.dealer.deal_card())
                    self.round_counter += 1
            elif self.round_counter == 2:
                self.stage = Stage.RIVER
                self.public_cards.append(self.dealer.deal_card())
                if len(self.players) == np.sum(players_in_bypass):
                    self.round_counter += 1

            self.round_counter += 1
            self.round.start_new_round(self.game_pointer)

        state = self.get_state(self.game_pointer)

        return state, self.game_pointer

    def get_state(self, player_id):
        ''' Return player's state

        Args:
            player_id (int): player id

        Returns:
            (dict): The state of the player
        '''
        self.dealer.pot = np.sum([player.in_chips for player in self.players])

        chips = [self.players[i].in_chips for i in range(self.num_players)]
        legal_actions = self.get_legal_actions()
        state = self.players[player_id].get_state(self.public_cards, chips, legal_actions)
        state['stakes'] = [self.players[i].remained_chips for i in range(self.num_players)]
        state['current_player'] = self.game_pointer
        state['pot'] = self.dealer.pot
        state['stage'] = self.stage
        return state

    def step_back(self):
        ''' Return to the previous state of the game

        Returns:
            (bool): True if the game steps back successfully
        '''
        if len(self.history) > 0:
            self.round, self.game_pointer, self.round_counter, self.dealer, self.public_cards, self.players = self.history.pop()
            return True
        return False

    def get_payoffs(self):
        ''' Return the payoffs of the game

        Returns:
            (list): Each entry corresponds to the payoff of one player
        '''
        hands = [p.hand + self.public_cards if p.status in (PlayerStatus.ALIVE, PlayerStatus.ALLIN) else None for p in self.players]
        chips_payoffs = self.judger.judge_game(self.players, hands)
        return chips_payoffs

    @staticmethod
    def get_action_num():
        ''' Return the number of applicable actions

        Returns:
            (int): The number of actions. There are 6 actions (call, raise_half_pot, raise_pot, all_in, check and fold)
        '''
        return len(Action)


#if __name__ == "__main__":
#    game = NolimitholdemGame()
#
#    while True:
#        print('New Game')
#        state, game_pointer = game.init_game()
#        print(game_pointer, state)
#        i = 1
#        while not game.is_over():
#            i += 1
#            legal_actions = game.get_legal_actions()
#            # if i == 3:
#            #     print('Step back')
#            #     print(game.step_back())
#            #     game_pointer = game.get_player_id()
#            #     print(game_pointer)
#            #     legal_actions = game.get_legal_actions()
#
#            action = np.random.choice(legal_actions)
#            # action = input()
#            # if action != 'call' and action != 'fold' and action != 'check':
#            #     action = int(action)
#            print(game_pointer, action, legal_actions)
#            state, game_pointer = game.step(action)
#            print(game_pointer, state)
#
#        print(game.get_payoffs())



'''
    Project: Gui Gin Rummy
    File name: env_thread.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from typing import List

import threading
import time

from rlcard.envs.gin_rummy import GinRummyEnv

from rlcard.games.gin_rummy.utils.move import GinRummyMove, ScoreSouthMove

from ..gin_rummy_human_agent import HumanAgent

from . import utils
from . import status_messaging


class EnvThread(threading.Thread):

    def __init__(self, gin_rummy_env: GinRummyEnv, game_canvas: 'GameCanvas'):
        super().__init__()
        self.gin_rummy_env = gin_rummy_env
        self.game_canvas = game_canvas
        self.setDaemon(daemonic=True)
        self.name = "WorkerGinRummyEnvironment"
        self.mark = 0
        self.is_stopped = False

    @property
    def moves(self) -> List[GinRummyMove]:
        return self.gin_rummy_env.game.round.move_sheet

    #
    #   Producer of action_ids
    #

    def is_action_id_available(self) -> bool:
        move_sheet = self.gin_rummy_env.game.round.move_sheet
        move_count = len(move_sheet)
        return self.mark < move_count

    #
    #   Consumer of action_ids
    #

    def get_waiting_player_id(self) -> int or None:  # FIXME: rework this
        waiting_player_id = None  # type: int or None
        is_action_id_available = self.is_action_id_available()
        if not is_action_id_available:
            for player_id, agent in enumerate(self.gin_rummy_env.agents):
                if isinstance(agent, HumanAgent) and agent.is_choosing_action_id:
                    waiting_player_id = player_id
        return waiting_player_id

    #
    #   Thread methods
    #

    def stop(self):
        self.is_stopped = True
        south_agent = self.gin_rummy_env.agents[1]
        if isinstance(south_agent, HumanAgent):
            if south_agent.is_choosing_action_id:
                legal_actions = south_agent.state['legal_actions']
                if legal_actions:
                    south_agent.chosen_action_id = legal_actions[0]
            self.gin_rummy_env.game.round.is_over = True

    def run(self) -> None:
        self.game_canvas.game_canvas_updater.apply_canvas_updates()
        _, payoffs = self.gin_rummy_env.run(is_training=False)

        # determine whether game completed all moves
        is_game_complete = False
        move_sheet = self.gin_rummy_env.game.round.move_sheet
        if move_sheet and isinstance(move_sheet[-1], ScoreSouthMove):
            is_game_complete = True

        # handle case where moves still need to be processed
        if is_game_complete:
            while not self.is_stopped and self.gin_rummy_env.game.round.is_over and self.mark < len(self.moves):
                time.sleep(0.1)  # FIXME: provide a timeout ???

        # on normal completion: update score_pads
        if not self.is_stopped and self.gin_rummy_env.game.round.is_over and is_game_complete:
            status_messaging.show_game_over_message(game=self.gin_rummy_env.game, game_canvas=self.game_canvas)
            for player_id in range(2):  # update score_pad for both players
                payoff = payoffs[player_id]
                if isinstance(payoffs, float):
                    text = "{:.2f}".format(payoff)
                else:
                    text = "{}".format(payoff)
                self.game_canvas.score_pad_cells[player_id].configure(text=text)

        if utils.is_debug():
            print("EnvThread finished")



import numpy as np
from copy import deepcopy

from rlcard.games.mahjong import Dealer
from rlcard.games.mahjong import Player
from rlcard.games.mahjong import Round
from rlcard.games.mahjong import Judger

class MahjongGame(object):

    def __init__(self, allow_step_back=False):
        '''Initialize the class MajongGame
        '''
        self.allow_step_back = allow_step_back
        self.np_random = np.random.RandomState()
        self.num_players = 4

    def init_game(self):
        ''' Initialilze the game of Mahjong

        This version supports two-player Mahjong

        Returns:
            (tuple): Tuple containing:

                (dict): The first state of the game
                (int): Current player's id
        '''
        # Initialize a dealer that can deal cards
        self.dealer = Dealer(self.np_random)

        # Initialize four players to play the game
        self.players = [Player(i, self.np_random) for i in range(self.num_players)]

        self.judger = Judger(self.np_random)
        self.round = Round(self.judger, self.dealer, self.num_players, self.np_random)

        # Deal 13 cards to each player to prepare for the game
        for player in self.players:
            self.dealer.deal_cards(player, 13)

        # Save the hisory for stepping back to the last state.
        self.history = []

        self.dealer.deal_cards(self.players[self.round.current_player], 1)
        state = self.get_state(self.round.current_player)
        self.cur_state = state
        return state, self.round.current_player

    def step(self, action):
        ''' Get the next state

        Args:
            action (str): a specific action. (call, raise, fold, or check)

        Returns:
            (tuple): Tuple containing:

                (dict): next player's state
                (int): next plater's id
        '''
        # First snapshot the current state
        if self.allow_step_back:
            hist_dealer = deepcopy(self.dealer)
            hist_round = deepcopy(self.round)
            hist_players = deepcopy(self.players)
            self.history.append((hist_dealer, hist_players, hist_round))
        self.round.proceed_round(self.players, action)
        state = self.get_state(self.round.current_player)
        self.cur_state = state
        return state, self.round.current_player

    def step_back(self):
        ''' Return to the previous state of the game

        Returns:
            (bool): True if the game steps back successfully
        '''
        if not self.history:
            return False
        self.dealer, self.players, self.round = self.history.pop()
        return True

    def get_state(self, player_id):
        ''' Return player's state

        Args:
            player_id (int): player id

        Returns:
            (dict): The state of the player
        '''
        state = self.round.get_state(self.players, player_id)
        return state

    @staticmethod
    def get_legal_actions(state):
        ''' Return the legal actions for current player

        Returns:
            (list): A list of legal actions
        '''
        if state['valid_act'] == ['play']:
            state['valid_act'] = state['action_cards']
            return state['action_cards']
        else:
            return state['valid_act']

    @staticmethod
    def get_action_num():
        ''' Return the number of applicable actions

        Returns:
            (int): The number of actions. There are 4 actions (call, raise, check and fold)
        '''
        return 38

    def get_player_num(self):
        ''' return the number of players in Mahjong

        returns:
            (int): the number of players in the game
        '''
        return self.num_players

    def get_player_id(self):
        ''' return the id of current player in Mahjong

        returns:
            (int): the number of players in the game
        '''
        return self.round.current_player

    def is_over(self):
        ''' Check if the game is over

        Returns:
            (boolean): True if the game is over
        '''
        win, player, _ = self.judger.judge_game(self)
        #pile =[sorted([c.get_str() for c in s ]) for s in self.players[player].pile if self.players[player].pile != None]
        #cards = sorted([c.get_str() for c in self.players[player].hand])
        #count = len(cards) + sum([len(p) for p in pile])
        self.winner = player
        #print(win, player, players_val)
        #print(win, self.round.current_player, player, cards, pile, count)
        return win

## For test
#if __name__ == '__main__':
#    import time
#    np.random.seed(2)
#    start = time.time()
#    game = MahjongGame()
#    for _ in range(100000):
#        #print('*****init game*****')
#        state, button = game.init_game()
#        i = 0
#        while not game.is_over():
#            i += 1
#            legal_actions = game.get_legal_actions(state)
#            action = np.random.choice(legal_actions)
#            flag=0
#            #if len(legal_actions) < 3:
#            #    flag=1
#            #    print("Before:", state)
#            #    print(game.round.current_player, action)
#            state, button = game.step(action)
#            #if action != 'stand' and flag==1:
#            #    print("After:", state)
#            #    print(state, game.round.current_player)
#            #    exit()
#            #print(button, state)
#        winner_hand = [c.get_str() for c in game.players[game.winner].hand]
#        winnder_pile = [[c.get_str() for c in s] for s in game.players[game.winner].pile]
#        print(_, len(game.dealer.deck), game.winner, winner_hand, winnder_pile)
#        if game.winner != -1:
#            exit()
#    end = time.time()
#    print(end-start)



'''
    Project: Gui Gin Rummy
    File name: configurations.py
    Author: William Hale
    Date created: 3/14/2020
'''

import os

from configparser import ConfigParser

#
#   Gin Rummy parameters
#

GOING_OUT_DEADWOOD_COUNT = 10

#
#   RLCard Gin Rummy parameters
#

MAX_DRAWN_CARD_COUNT = 52

DISCARD_PILE_TAG = "discard_pile"
STOCK_PILE_TAG = "stock_pile"
NORTH_HELD_PILE_TAG = "north_held_pile"
SOUTH_HELD_PILE_TAG = "south_held_pile"
PLAYER_HELD_PILE_TAGS = [NORTH_HELD_PILE_TAG, SOUTH_HELD_PILE_TAG]

DRAWN_TAG = "drawn"
JOGGED_TAG = "jogged"
SELECTED_TAG = "selected"

SCORE_PLAYER_0_ACTION_ID = 0
SCORE_PLAYER_1_ACTION_ID = 1
DRAW_CARD_ACTION_ID = 2
PICK_UP_DISCARD_ACTION_ID = 3
DECLARE_DEAD_HAND_ACTION_ID = 4
GIN_ACTION_ID = 5
DISCARD_ACTION_ID = 6
KNOCK_ACTION_ID = DISCARD_ACTION_ID + 52

#
#   Not User Modifiable Options
#

IS_KEEP_TURN_WHEN_DISCARDING_CARD_PICKED_UP = False  # TODO: make True the default value

#
#   User Modifiable Options
#

config_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'game_options.ini')  # Note this

config = ConfigParser()
found = config.read(config_path)

# settings section
settings_section = "settings"
show_status_messages_option = "show_status_messages"
warning_as_option = 'warning_as'
game_background_color_option = 'game_background_color'
window_size_factor_option = 'window_size_factor'
is_show_tips_option = "is_show_tips"
is_debug_option = "is_debug"

SHOW_STATUS_MESSAGES = config.get(section=settings_section, option=show_status_messages_option, fallback="verbose")
WARNINGS_AS = config.get(section=settings_section, option=warning_as_option, fallback="alert_messages")
GAME_BACKGROUND_COLOR = config.get(section=settings_section, option=game_background_color_option, fallback="#007F00")
WINDOW_SIZE_FACTOR = config.getint(section=settings_section, option=window_size_factor_option, fallback=75)
IS_SHOW_TIPS = config.getboolean(section=settings_section, option=is_show_tips_option, fallback=True)
# Note: IS_DEBUG always starts off as False; must explicitly update via preference window
# IS_DEBUG = config.getboolean(section=settings_section, option=is_debug_option, fallback=False)
IS_DEBUG = False



'''
    Project: Gui Gin Rummy
    File name: handling_tap_stock_pile.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from .canvas_item import CanvasItem

from . import configurations
from . import info_messaging
from . import utils


def handle_tap_stock_pile(hit_item: CanvasItem, game_canvas: 'GameCanvas'):  # hit_item is source
    # Normal case is can_draw_from_stock_pile.
    # hit_item must not be drawn.
    # hit_item must be top card of stock_pile.
    # reset top card of discard pile if drawn.
    # reset all selected cards in held_pile
    player_id = game_canvas.current_player_id
    if game_canvas.query.is_game_over():
        pass
    elif game_canvas.query.can_discard_card(player_id=player_id):
        pass
    elif game_canvas.query.can_declare_dead_hand(player_id=player_id):
        pass
    elif game_canvas.query.can_draw_from_stock_pile(player_id=player_id):
        current_player_id = game_canvas.current_player_id
        hit_item_tags = hit_item.get_tags()
        if configurations.DRAWN_TAG not in hit_item_tags:
            top_stock_pile_item_id = game_canvas.getter.get_top_stock_pile_item_id()
            if hit_item == top_stock_pile_item_id:
                utils.toggle_stock_pile_item_selected(game_canvas)
                # reset drawn top card of discard_pile if needed
                top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
                top_discard_pile_item_tags = game_canvas.getter.get_tags(top_discard_pile_item_id)
                if configurations.DRAWN_TAG in top_discard_pile_item_tags:
                    utils.toggle_discard_pile_item_selected(game_canvas=game_canvas)
                # reset selected cards of held_pile of current_player
                held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=current_player_id)
                for held_pile_item_id in held_pile_item_ids:
                    held_pile_item_tags = game_canvas.getter.get_tags(item_id=held_pile_item_id)
                    if configurations.SELECTED_TAG in held_pile_item_tags:
                        held_pile_item = game_canvas.canvas_item_by_item_id.get(held_pile_item_id, None)
                        if held_pile_item:
                            utils.toggle_held_pile_item_selected(item=held_pile_item, game_canvas=game_canvas)
                # remove info_message if south player
                if player_id == 1:
                    info_messaging.blank_info_message_label(game_canvas=game_canvas)


import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

extras = {
    'torch': ['torch>=1.3'],
    'tensorflow': ['tensorflow>=1.14,<2.0']
}

def _get_version():
    with open('rlcard/__init__.py') as f:
        for line in f:
            if line.startswith('__version__'):
                g = {}
                exec(line, g)
                return g['__version__']
        raise ValueError('`__version__` not defined')

VERSION = _get_version()

setuptools.setup(
    name="rlcard",
    version=VERSION,
    author="Data Analytics at Texas A&M (DATA) Lab",
    author_email="khlai037@tamu.edu",
    description="A Toolkit for Reinforcement Learning in Card Games",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/datamllab/rlcard",
    keywords=["Reinforcement Learning", "game", "RL", "AI"],
    packages=setuptools.find_packages(exclude=('tests',)),
    package_data={
        'rlcard': ['models/pretrained/leduc_holdem_nfsp/*',
                   'models/pretrained/leduc_holdem_cfr/*',
                   'models/pretrained/leduc_holdem_nfsp_pytorch/*',
                   'games/uno/jsondata/action_space.json',
                   'games/limitholdem/card2index.json',
                   'games/leducholdem/card2index.json',
                   'games/doudizhu/jsondata/*',
                   'games/uno/jsondata/*',
                   'games/simpledoudizhu/jsondata/*',
                   'agents/gin_rummy_human_agent/gui_cards/*',
                   'agents/gin_rummy_human_agent/gui_cards/cards_png/*',
                   'agents/gin_rummy_human_agent/gui_gin_rummy/*'
                   ]},
    install_requires=[
        'numpy>=1.16.3',
        'matplotlib>=3.0',
        'pillow>=5.2.0',
        'termcolor',
        'packaging',
    ],
    extras_require=extras,
    requires_python='>=3.5',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



'''
    Project: Gui Gin Rummy
    File name: handling_tap_held_pile.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError

from .player_type import PlayerType
from .canvas_item import CanvasItem

from . import configurations
from . import info_messaging
from . import utils


def handle_tap_held_pile(hit_item: CanvasItem, game_canvas: 'GameCanvas'):
    hit_item_tags = hit_item.get_tags()
    if game_canvas.held_pile_tags[0] in hit_item_tags:
        player_id = 0
    elif game_canvas.held_pile_tags[1] in hit_item_tags:
        player_id = 1
    else:
        raise GinRummyProgramError("handle_tap_held_pile: unknown held_pile.")
    player_is_human = game_canvas.player_types[player_id] is PlayerType.human_player
    can_draw_from_stock_pile = game_canvas.query.can_draw_from_stock_pile(player_id=player_id)
    can_draw_from_discard_pile = game_canvas.query.can_draw_from_discard_pile(player_id=player_id)
    is_game_over = game_canvas.query.is_game_over()
    if is_game_over:
        pass
    elif game_canvas.query.can_discard_card(player_id=player_id):  # hit_item is source
        if player_is_human:
            utils.toggle_held_pile_item_selected(item=hit_item, game_canvas=game_canvas)
    elif can_draw_from_stock_pile or can_draw_from_discard_pile:  # hit_item is target
        drawn_card_item_id = None
        drawn_card_item_tag = None
        if not drawn_card_item_id and can_draw_from_stock_pile:
            top_stock_pile_item_id = game_canvas.getter.get_top_stock_pile_item_id()
            top_stock_pile_item_tags = game_canvas.getter.get_tags(top_stock_pile_item_id)
            if configurations.DRAWN_TAG in top_stock_pile_item_tags:
                drawn_card_item_id = top_stock_pile_item_id
                drawn_card_item_tag = configurations.STOCK_PILE_TAG
        if not drawn_card_item_id and can_draw_from_discard_pile:
            top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
            top_discard_pile_item_tags = game_canvas.getter.get_tags(top_discard_pile_item_id)
            if configurations.DRAWN_TAG in top_discard_pile_item_tags:
                drawn_card_item_id = top_discard_pile_item_id
                drawn_card_item_tag = configurations.DISCARD_PILE_TAG
        if drawn_card_item_id:
            if player_id == 1:  # remove info_message if south player
                info_messaging.blank_info_message_label(game_canvas=game_canvas)
            game_canvas.post_doing_action.post_do_get_card_action(player_id=player_id,
                                                                  drawn_card_item_id=drawn_card_item_id,
                                                                  hit_item_id=hit_item.item_id,
                                                                  drawn_card_item_tag=drawn_card_item_tag)
        else:
            utils.toggle_held_pile_item_selected(item=hit_item, game_canvas=game_canvas)
    else:
        if player_is_human:
            utils.toggle_held_pile_item_selected(item=hit_item, game_canvas=game_canvas)  # arranging hand



'''
    Project: Gui Gin Rummy
    File name: gin_rummy_human_agent.py
    Author: William Hale
    Date created: 3/14/2020
'''

import time

from rlcard.games.gin_rummy.utils.action_event import ActionEvent
from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError


class HumanAgent(object):
    ''' A human agent for Gin Rummy. It can be used to play against trained models.
    '''

    def __init__(self, action_num):
        ''' Initialize the human agent

        Args:
            action_num (int): the size of the output action space
        '''
        self.use_raw = True
        self.action_num = action_num
        self.is_choosing_action_id = False
        self.chosen_action_id = None  # type: int or None
        self.state = None

    def step(self, state):
        ''' Human agent will display the state and make decisions through interfaces

        Args:
            state (dict): A dictionary that represents the current state

        Returns:
            action (int): The action decided by human
        '''
        if self.is_choosing_action_id:
            raise GinRummyProgramError("self.is_choosing_action_id must be False.")
        if self.state is not None:
            raise GinRummyProgramError("self.state must be None.")
        if self.chosen_action_id is not None:
            raise GinRummyProgramError("self.chosen_action_id={} must be None.".format(self.chosen_action_id))
        self.state = state
        self.is_choosing_action_id = True
        while not self.chosen_action_id:
            time.sleep(0.001)
        if self.chosen_action_id is None:
            raise GinRummyProgramError("self.chosen_action_id cannot be None.")
        chosen_action_event = ActionEvent.decode_action(action_id=self.chosen_action_id)
        self.state = None
        self.is_choosing_action_id = False
        self.chosen_action_id = None
        return chosen_action_event

    def eval_step(self, state):
        ''' Predict the action given the current state for evaluation. The same to step here.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action (int): the action predicted (randomly chosen) by the random agent
            probabilities (list): The list of action probabilities
        '''
        return self.step(state), []



'''
    Project: Gui Gin Rummy
    File name: card_image.py
    Author: William Hale
    Date created: 3/14/2020
'''

import os
from PIL import Image, ImageTk, ImageDraw

image_dir = os.path.abspath(os.path.dirname(__file__))

ranks = ['A', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K']
suits = ['C', 'D', 'H', 'S']


def long_rank_name_for(rank: str) -> str:
    rank_exceptions = {'A': 'ace', 'T': '10', 'J': 'jack', 'Q': 'queen', 'K': 'king'}
    result = rank if rank not in rank_exceptions.keys() else rank_exceptions[rank]
    return result


def long_suit_name_for(suit: str) -> str:
    known_suit_name_by_suit = {'C': 'clubs', 'D': 'diamonds', 'H': 'hearts', 'S': 'spades'}
    result = suit if suit not in known_suit_name_by_suit else known_suit_name_by_suit[suit]
    return result


long_rank_name_by_rank = {(rank, long_rank_name_for(rank)) for rank in ranks}
long_suit_name_by_suit = {(suit, long_suit_name_for(suit)) for suit in suits}


def get_card_filename(rank: str, suit: str) -> str:
    long_rank_name = long_rank_name_for(rank)
    long_suit_name = long_suit_name_for(suit)
    card_file_name = "{}/cards_png/{}_of_{}@2x.png".format(image_dir, long_rank_name, long_suit_name)
    return card_file_name


def get_card_back_image(scale_factor: float):
    card_filename = "{}/back@2x.png".format(image_dir)
    image = Image.open(card_filename)
    image_width, image_height = image.size
    card_scale_factor = 0.25 * scale_factor * 0.666
    card_image_width = int(image_width * card_scale_factor)
    card_image_height = int(image_height * card_scale_factor)
    image = image.resize((card_image_width, card_image_height), Image.ANTIALIAS)
    return image


class CardImage(ImageTk.PhotoImage):

    def __init__(self, rank: str, suit: str, scale_factor: float):
        self.scale_factor = scale_factor
        self.face_up = True
        card_filename = get_card_filename(rank=rank, suit=suit)
        image = Image.open(card_filename)
        image_width, image_height = image.size
        card_scale_factor = 0.25 * self.scale_factor
        card_image_width = int(image_width * card_scale_factor)
        card_image_height = int(image_height * card_scale_factor)
        image = image.resize((card_image_width, card_image_height), Image.ANTIALIAS)
        super().__init__(image=image, name="{}{}".format(rank, suit))


class CardBackImage(ImageTk.PhotoImage):

    def __init__(self, scale_factor: float):
        self.scale_factor = scale_factor
        card_filename = get_card_filename(rank='A', suit='C')  # want to use AC which has a lot of white space
        image = Image.open(card_filename)
        image_width, image_height = image.size
        card_scale_factor = 0.25 * self.scale_factor
        card_image_width = int(image_width * card_scale_factor)
        card_image_height = int(image_height * card_scale_factor)
        image = image.resize((card_image_width, card_image_height), Image.ANTIALIAS)
        draw = ImageDraw.Draw(image)
        shape = [3 * self.scale_factor,
                 3 * self.scale_factor,
                 card_image_width - 3 * self.scale_factor,
                 card_image_height - 3 * self.scale_factor]
        draw.rectangle(shape, fill='white', outline=None)
        # paste back image
        card_back_image = get_card_back_image(scale_factor=1.08 * self.scale_factor)
        width = card_back_image.width
        height = card_back_image.height
        crop_left = int(10 * self.scale_factor)
        crop_top = int(10 * self.scale_factor)
        crop_right = width - int(10 * self.scale_factor)
        crop_bottom = height - int(10 * self.scale_factor)
        card_back_image = card_back_image.crop((crop_left, crop_top, crop_right, crop_bottom))
        image.paste(card_back_image, (int(6 * self.scale_factor), int(7 * self.scale_factor)))
        super().__init__(image=image, name="cardBackImage")


class BlankCardImage(ImageTk.PhotoImage):  # FIXME: not used

    def __init__(self, rank: str, suit: str, scale_factor: float):
        self.scale_factor = scale_factor
        card_filename = get_card_filename(rank=rank, suit=suit)
        image = Image.open(card_filename)
        image_width, image_height = image.size
        card_scale_factor = 0.25 * self.scale_factor
        card_image_width = int(image_width * card_scale_factor)
        card_image_height = int(image_height * card_scale_factor)
        image = image.resize((card_image_width, card_image_height), Image.ANTIALIAS)
        # image.putalpha(128)

        # draw = ImageDraw.Draw(image)
        # shape = [(3 * self.scale_factor, 3 * self.scale_factor),
        #          (card_image_width - 3 * self.scale_factor, card_image_height - 3 * self.scale_factor)]
        # draw.rectangle(shape, fill="gray", outline=None)
        # image.putalpha(128)
        draw = ImageDraw.Draw(image)
        shape = [(0, 0), (card_image_width, card_image_height)]
        draw.rectangle(shape, fill="black", outline=int(1 * self.scale_factor))
        # image.putalpha(128)

        super().__init__(image=image, name="blankCardImage")


class BorderLessCardImage(ImageTk.PhotoImage):  # FIXME: not used

    def __init__(self, rank: str, suit: str, scale_factor: float):
        self.scale_factor = scale_factor
        card_filename = get_card_filename(rank=rank, suit=suit)
        image = Image.open(card_filename)
        image_width, image_height = image.size
        card_scale_factor = 0.25 * self.scale_factor
        card_image_width = int(image_width * card_scale_factor)
        card_image_height = int(image_height * card_scale_factor)
        image = image.resize((card_image_width, card_image_height), Image.ANTIALIAS)

        draw = ImageDraw.Draw(image)
        shape = [(0, 0), (card_image_width, card_image_height)]
        draw.rectangle(shape, fill="white", outline=int(1 * scale_factor))
        for i in range(4):
            top = 0
            right = 0
            left = card_image_width
            bottom = card_image_height
            shapes = []
            shape_size = 3.5 * scale_factor
            shapes.append((right, top, right + shape_size, top + shape_size))
            shapes.append((left - shape_size - 1 * scale_factor, top, left + shape_size, top + shape_size))
            shapes.append((left - shape_size - 0 * scale_factor, bottom - shape_size, left, bottom))
            shapes.append((right, bottom - shape_size, right + shape_size, bottom))
            ImageDraw.Draw(image).rectangle(shapes[i], fill="white")

        super().__init__(image=image, name="{}{}".format(rank, suit))



'''
    Project: Gui Gin Rummy
    File name: game_frame.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_app import GameApp

import tkinter as tk

from . import configurations
from . import starting_new_game
from . import utils

from .game_canvas import GameCanvas


class GameFrame(tk.Frame):

    def __init__(self, root: tk.Tk, game_app: 'GameApp'):
        # Set window_title.
        window_title = "Gin Rummy"
        # Set base size.
        # Only the aspect ratio matters for frame size.
        base_height = 750  # type: int
        base_width = 1000  # type: int
        # Set scale_factor_multiplier.
        #   You can change scale_factor_multiplier to increase or decrease the size of the cards,
        #   leaving the size of the the frame the same.
        scale_factor_multiplier = 1.5  # type: float  # Note this: change to scale card size

        preference_window_size_factor = configurations.WINDOW_SIZE_FACTOR
        try:
            window_size_factor = preference_window_size_factor / 100  # type: float
        except ValueError:
            window_size_factor = 0.75  # type: float

        # Get working screen size.
        screenheight_fudge_factor = 0.9  # type: float  # take into account space used by system menus, task bars, etc
        screenwidth = root.winfo_screenwidth()  # screen's width in pixels
        screenheight = root.winfo_screenheight()  # screen's height in pixels
        working_screen_width = int(screenwidth)  # type: int
        working_screen_height = int(screenheight * screenheight_fudge_factor)

        # Determine max frame size.
        min_ratio = min(working_screen_width / base_width, working_screen_height / base_height)  # type: float
        max_frame_width = base_width * min_ratio
        max_frame_height = base_height * min_ratio

        # Determine window rectangle.
        window_width = int(max_frame_width * window_size_factor)  # type: int
        window_height = int(max_frame_height * window_size_factor)  # type: int
        window_x = int((working_screen_width - window_width) / 2)  # type: int
        window_y = int((working_screen_height - window_height) / 2)  # type: int

        super().__init__(root, width=window_width, height=window_height)

        # Set up frame.
        self.root = root
        self.root.title(window_title)
        self.window_size_factor = max(min(window_size_factor, 1), 0.5)
        self.root.geometry("{}x{}+{}+{}".format(window_width, window_height, window_x, window_y))

        # Set scale_factor for card size.
        self.scale_factor = scale_factor_multiplier * window_size_factor  # type: float

        # Create canvas.
        self.game_canvas = GameCanvas(parent=self, window_width=window_width,
                                      window_height=window_height,
                                      scale_factor=self.scale_factor,
                                      game_app=game_app)
        self.pack()
        self.start_new_game()

        if utils.is_debug():
            self.update()
            print("screen size is {}x{}".format(screenwidth, screenheight))
            print("working screen size is {}x{}".format(working_screen_width, working_screen_height))
            print("(base_width, base_height) is ({}, {})".format(base_width, base_height))
            print("(window_width, window_height) is ({}, {})".format(window_width, window_height))
            print("window size is ({}, {})".format(self.winfo_width(), self.winfo_height()))
            print("window_size_factor={}".format(window_size_factor))
            print("scale_factor={}".format(self.scale_factor))

    def start_new_game(self):
        starting_new_game.start_new_game(game_canvas=self.game_canvas)

    def update_configurations(self):
        self.game_canvas.update_configurations()

    def update_configuration_game_background_color(self, background_color):
        self.game_canvas.update_configuration_game_background_color(background_color=background_color)



from rlcard.games.leducholdem.dealer import LeducholdemDealer as Dealer
from rlcard.games.leducholdem.judger import LeducholdemJudger as Judger
from rlcard.games.leducholdem.player import LeducholdemPlayer as Player
from rlcard.games.leducholdem.round import LeducholdemRound as Round
from rlcard.games.leducholdem.game import LeducholdemGame as Game




'''
    Project: Gui Gin Rummy
    File name: starting_new_game.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

import tkinter as tk

from ..gin_rummy_human_agent import HumanAgent

from . import configurations
from . import info_messaging
from . import utils

from .env_thread import EnvThread

import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils
from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError


def start_new_game(game_canvas: 'GameCanvas'):
    info_messaging.blank_info_message_label(game_canvas=game_canvas)
    if game_canvas.game_canvas_updater.env_thread and game_canvas.game_canvas_updater.env_thread.is_alive():
        game_canvas.game_canvas_updater.env_thread.stop()  # FIXME: complex stopping of threads; simplify ???
        while game_canvas.game_canvas_updater.env_thread.is_alive() or not game_canvas.game_canvas_updater.is_stopped:
            game_canvas.update()  # yield time to other threads and to main thread
        if game_canvas.game_canvas_updater.env_thread.is_alive():
            raise GinRummyProgramError("env_thread did not stop.")
    _reset_game_canvas(game_canvas=game_canvas)
    # make new gin_rummy_env
    gin_rummy_env = game_canvas.game_app.make_gin_rummy_env()
    gin_rummy_env.game.settings.going_out_deadwood_count = configurations.GOING_OUT_DEADWOOD_COUNT  # Note this
    gin_rummy_env.game.settings.max_drawn_card_count = configurations.MAX_DRAWN_CARD_COUNT  # Note this
    # start thread
    if utils.is_debug():
        south_agent = gin_rummy_env.agents[1]
        if isinstance(south_agent, HumanAgent):
            if south_agent.state is not None:
                raise GinRummyProgramError("south_agent.state must be None.")
            if south_agent.is_choosing_action_id is True:
                raise GinRummyProgramError("south_agent.is_choosing_action_id must be False.")
            if south_agent.chosen_action_id is not None:
                raise GinRummyProgramError("south_agent.chosen_action_id={} must be None.".format(south_agent.chosen_action_id))
    game_canvas.game_canvas_updater.env_thread = EnvThread(gin_rummy_env=gin_rummy_env, game_canvas=game_canvas)
    game_canvas.game_canvas_updater.env_thread.start()  # Note this: start env background thread


def _reset_game_canvas(game_canvas: 'GameCanvas'):
    game_canvas.dealer_id = 0  # Note: has no effect; will be set later to correct value
    game_canvas.game_canvas_updater.pending_human_action_ids = []
    game_canvas.game_canvas_updater.busy_body_id = None
    game_canvas.game_canvas_updater.is_stopped = False
    for player_id in range(2):
        game_canvas.score_pad_cells[player_id].configure(text="")
    for card_id in range(52):
        card_item_id = game_canvas.card_item_ids[card_id]
        game_canvas.itemconfigure(card_item_id, tags=[])
    game_canvas.update()  # seems to be needed: scores not cleared if 'new game' chosen from menu with no mouse move
    if utils.is_debug():
        for card_id in range(52):
            card_item_id = game_canvas.card_item_ids[card_id]
            tags = game_canvas.gettags(card_item_id)
            if tags:
                raise GinRummyProgramError("tags must be None.")


def show_new_game(game_canvas: 'GameCanvas'):
    game = game_canvas.game_canvas_updater.env_thread.gin_rummy_env.game
    game_canvas.dealer_id = game.round.dealer_id
    dealer = game.round.dealer
    shuffled_deck = dealer.shuffled_deck
    # Deal cards to players
    if utils.is_debug():
        for card_id in range(52):
            card_item_id = game_canvas.card_item_ids[card_id]
            tags = game_canvas.gettags(card_item_id)
            if tags:
                raise GinRummyProgramError("tags must be None.")
    for i in range(2):
        player_id = (game_canvas.dealer_id + 1 + i) % 2
        anchor_x, anchor_y = game_canvas.player_held_pile_anchors[player_id]
        face_up = True
        # Note: cannot use card_ids = [gin_rummy_utils.get_card_id(card) for card in game.round.players[player_id].hand]
        if not game_canvas.is_treating_as_human(player_id=player_id):
            face_up = False
        if i == 0:
            dealt_cards = list(shuffled_deck[-11:])
        else:
            dealt_cards = list(shuffled_deck[-21:-11])
        card_ids = [gin_rummy_utils.get_card_id(card=card) for card in dealt_cards]
        card_ids = sorted(card_ids, reverse=True, key=utils.gin_rummy_sort_order_id)
        for card_id in card_ids:
            card_item_id = game_canvas.card_item_ids[card_id]
            if utils.is_debug() or True:
                tags = game_canvas.gettags(card_item_id)
                if tags:
                    raise GinRummyProgramError("tags must be None.")
            game_canvas.tag_raise(card_item_id)  # note this
            game_canvas.itemconfig(card_item_id, tag=game_canvas.held_pile_tags[player_id])
            utils.set_card_id_face_up(card_id=card_id, face_up=face_up, game_canvas=game_canvas)
            utils.move_to(card_item_id, anchor_x, anchor_y, parent=game_canvas)
            game_canvas.itemconfigure(card_item_id, state=tk.NORMAL)
            anchor_x += game_canvas.held_pile_tab
    # Deal stockpile cards
    stock_pile_x, stock_pile_y = game_canvas.stock_pile_anchor
    # Note: cannot use stock_pile = [gin_rummy_utils.get_card_id(card) for card in game.round.dealer.stock_pile]
    stock_pile = list(shuffled_deck[:31])
    stock_pile_card_ids = [gin_rummy_utils.get_card_id(card) for card in stock_pile]
    for card_id in stock_pile_card_ids:
        card_item_id = game_canvas.card_item_ids[card_id]
        if utils.is_debug():
            tags = game_canvas.gettags(card_item_id)
            if tags:
                raise GinRummyProgramError("tags must be None.")
        game_canvas.tag_raise(card_item_id)  # note this
        game_canvas.itemconfig(card_item_id, tag=configurations.STOCK_PILE_TAG)
        utils.set_card_id_face_up(card_id=card_id, face_up=False, game_canvas=game_canvas)
        utils.move_to(card_item_id, stock_pile_x, stock_pile_y, parent=game_canvas)
        game_canvas.itemconfigure(card_item_id, state=tk.NORMAL)
        stock_pile_x += game_canvas.stock_pile_tab
    # update canvas
    game_canvas.update()  # seems to be needed: scores not cleared if 'new game' chosen from menu with no mouse move
    if utils.is_debug():
        settings = game.settings
        settings.print_settings()



from rlcard.games.limitholdem import Judger

class NolimitholdemJudger(Judger):
    ''' The Judger class for Texas Hold'em
    '''

    def __init__(self, np_random):
        ''' Initialize a judger class
        '''
        super(NolimitholdemJudger, self).__init__(np_random)



'''
    Project: Gui Gin Rummy
    File name: preferences_window.py
    Author: William Hale
    Date created: 3/14/2020
'''

from tkinter import *
import tkinter.colorchooser as colorchooser
from configparser import ConfigParser

from . import configurations
from .configurations import config_path
from .configurations import settings_section
from .configurations import show_status_messages_option, warning_as_option, game_background_color_option
from .configurations import window_size_factor_option
from .configurations import is_show_tips_option
from .configurations import is_debug_option


class PreferencesWindow(object):

    #
    # The game_background_color is updated immediately when returned from color dialog.
    # If the preference dialog is cancelled, the original game_background_color is restored.
    #

    # Note:
    #   The file game_options.ini must already exists with the settings sections existing at least.
    #   For this program, the line "[settings]" is the minimal that is required.

    def __init__(self, view):
        self.parent = view.master
        self.view = view

        # create preferences window
        self.pref_window = Toplevel(self.parent)
        self.pref_window.title("Gin Rummy Preferences")

        # fill preferences that cannot be obtained from their widgets
        self.game_background_color = configurations.GAME_BACKGROUND_COLOR
        self.original_game_background_color = self.game_background_color  # used if color dialog cancelled

        # create preferences pane (can be rearranged within same section)
        row = 0

        # Preference: Show status messages
        row += 1
        self.show_status_messages_listbox = self.create_choice_line(row=row, choice_name=show_status_messages_option)

        # Preference: Warnings as
        row += 1
        self.warnings_as_listbox = self.create_choice_line(row=row, choice_name=warning_as_option)

        # Preference: Game background color
        row += 1
        Label(self.pref_window, text="Game background color").grid(row=row, sticky=W, padx=5, pady=5)
        self.game_background_color_button = Button(self.pref_window, text='Select color')
        self.game_background_color_button.configure(command=self.set_game_background_color)
        self.game_background_color_button.grid(row=row, column=1, columnspan=2, sticky=W, padx=5, pady=5)

        # Preference: Window size
        row += 1
        Label(self.pref_window, text="Window size (percent)").grid(row=row, sticky=W, padx=5, pady=5)
        self.window_size_scale = Scale(self.pref_window, from_=50, to=100, orient=HORIZONTAL)
        self.window_size_scale.set(configurations.WINDOW_SIZE_FACTOR)
        self.window_size_scale.grid(row=row, column=1, columnspan=2, sticky=W, padx=5, pady=5)

        # Preference: is_show_tips
        row += 1
        self.is_show_tips = BooleanVar()
        self.is_show_tips.set(configurations.IS_SHOW_TIPS)
        self.is_show_tips_checkbutton = self.create_checkbutton(text="show tips", variable=self.is_show_tips)
        self.is_show_tips_checkbutton.grid(row=row, column=0, columnspan=3, sticky=W, padx=5, pady=5)

        # Preference: is_debug
        row += 1
        self.is_debug = BooleanVar()
        self.is_debug.set(configurations.IS_DEBUG)
        self.is_debug_checkbutton = self.create_checkbutton(text="is debug", variable=self.is_debug)
        self.is_debug_checkbutton.grid(row=row, column=0, columnspan=3, sticky=W, padx=5, pady=5)

        # create cancel and save buttons
        row += 1
        cancel_button = Button(self.pref_window, text="Cancel", command=self.on_cancel_button_clicked)
        cancel_button.grid(row=row, column=1, sticky=E, padx=5, pady=5)
        save_button = Button(self.pref_window, text="Save", command=self.on_save_button_clicked)
        save_button.grid(row=row, column=2, sticky=E, padx=5, pady=5)

        # wrap up
        self.pref_window.transient(self.parent)

    def set_game_background_color(self):  # store because color cannot be obtained from its widget
        self.game_background_color = colorchooser.askcolor(initialcolor=self.game_background_color)[-1]
        self.view.update_configuration_game_background_color(background_color=self.game_background_color)

    def on_save_button_clicked(self):
        self.set_new_values()
        self.pref_window.destroy()
        self.view.update_configurations()

    def set_new_values(self):
        config = ConfigParser()
        config.read(config_path)

        show_status_messages = self.show_status_messages_listbox.get(ACTIVE)
        configurations.SHOW_STATUS_MESSAGES = show_status_messages
        config.set(settings_section, show_status_messages_option, show_status_messages)

        warnings_as = self.warnings_as_listbox.get(ACTIVE)
        configurations.WARNINGS_AS = warnings_as
        config.set(settings_section, warning_as_option, warnings_as)

        configurations.GAME_BACKGROUND_COLOR = self.game_background_color
        config.set(settings_section, game_background_color_option, self.game_background_color)

        window_size_factor = self.window_size_scale.get()
        configurations.WINDOW_SIZE_FACTOR = window_size_factor
        config.set(settings_section, window_size_factor_option, str(window_size_factor))  # Note: using str

        is_show_tips = self.is_show_tips.get()
        configurations.IS_SHOW_TIPS = is_show_tips
        config.set(settings_section, is_show_tips_option, str(is_show_tips))  # Note: using str

        is_debug = self.is_debug.get()
        configurations.IS_DEBUG = is_debug
        config.set(settings_section, is_debug_option, str(is_debug))  # Note: using str

        with open(config_path, 'w') as config_file:
            config.write(config_file)

    def on_cancel_button_clicked(self):
        self.view.update_configuration_game_background_color(background_color=self.original_game_background_color)
        self.pref_window.destroy()

    def create_checkbutton(self, text: str, variable: BooleanVar):
        checkbutton = Checkbutton(self.pref_window, text=text, variable=variable)
        return checkbutton

    def create_listbox(self, row, column, columnspan, items):
        listbox = Listbox(self.pref_window, height=0, width=0)
        listbox.configure(exportselection=0)  # note this
        for item in items:
            listbox.insert(END, item)
        listbox.grid(row=row, column=column, columnspan=columnspan, sticky=W, padx=5, pady=5)
        return listbox

    def create_choice_line(self, row: int, choice_name: str):
        listbox = None
        items = None
        selected_item = None
        if choice_name == show_status_messages_option:
            Label(self.pref_window, text="Show status messages").grid(row=row, sticky=W, padx=5, pady=5)
            items = ["verbose", "brief", 'none']
            selected_item = configurations.SHOW_STATUS_MESSAGES
        elif choice_name == warning_as_option:
            Label(self.pref_window, text="Warnings as").grid(row=row, sticky=W, padx=5, pady=5)
            items = ["alert messages", "beeps"]
            selected_item = configurations.WARNINGS_AS
        if items is not None:
            listbox = self.create_listbox(row=row, column=1, columnspan=2, items=items)
            selected_item_index = 0
            if selected_item in items:
                selected_item_index = items.index(selected_item)
            listbox.selection_set(first=selected_item_index)
            listbox.activate(index=selected_item_index)
        return listbox



'''
    Project: Gui Gin Rummy
    File name: handling_tap_discard_pile.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from .canvas_item import CanvasItem

from . import configurations
from . import info_messaging
from . import starting_new_game
from . import utils

from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError


def handle_tap_discard_pile(hit_item: CanvasItem, game_canvas: 'GameCanvas'):
    player_id = game_canvas.current_player_id
    if game_canvas.query.is_game_over():
        starting_new_game.start_new_game(game_canvas)
    else:
        if game_canvas.query.can_draw_from_discard_pile(player_id=player_id):
            _handle_can_draw_from_discard_pile(hit_item=hit_item, game_canvas=game_canvas)
        elif game_canvas.query.can_discard_card(player_id=player_id):
            _handle_can_discard_card(hit_item=hit_item, game_canvas=game_canvas)
        else:
            pass  # FIXME: put warning message ???


#
#   Private methods
#

def _handle_can_draw_from_discard_pile(hit_item: CanvasItem, game_canvas: 'GameCanvas'):  # hit_item is source
    # hit_item must be top card of discard_pile.
    # Normal case is to toggle jog the top card of the discard pile.
    # Special case is to discard the card drawn from stockpile before being placed in held_pile.
    top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
    if hit_item.item_id == top_discard_pile_item_id:
        top_stock_pile_item_id = game_canvas.getter.get_top_stock_pile_item_id()
        top_stock_pile_item_tags = game_canvas.getter.get_tags(top_stock_pile_item_id)
        if configurations.DRAWN_TAG in top_stock_pile_item_tags:
            game_canvas.post_doing_action.post_do_discard_card_drawn_from_stock_pile_action(top_stock_pile_item_id)
        else:
            utils.toggle_discard_pile_item_selected(game_canvas=game_canvas)
    else:
        pass  # FIXME: put warning message ???


def _handle_can_discard_card(hit_item: CanvasItem, game_canvas: 'GameCanvas'):  # hit_item is target
    # hit_item must be top card of discard_pile or discard_pile_box_item.
    # exactly one held_pile_item must be selected.
    # The selected held_pile_item is discarded.
    # The remaining held_pile_items are fanned.
    top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
    if top_discard_pile_item_id:
        if not hit_item == top_discard_pile_item_id:
            raise GinRummyProgramError("hit_item must be top card of discard_pile.")
    else:
        if not hit_item == game_canvas.discard_pile_box_item:
            raise GinRummyProgramError("hit_item must be discard_pile_box_item.")
    current_player_id = game_canvas.current_player_id
    selected_held_pile_item_ids = game_canvas.getter.get_selected_held_pile_item_ids(player_id=current_player_id)
    if len(selected_held_pile_item_ids) == 1:
        if current_player_id == 1:  # remove info_message if south player
            info_messaging.blank_info_message_label(game_canvas=game_canvas)
        selected_held_pile_item_id = selected_held_pile_item_ids[0]
        game_canvas.post_doing_action.post_do_discard_action(player_id=current_player_id,
                                                             selected_held_pile_item_id=selected_held_pile_item_id)
    else:
        pass  # FIXME: put warning message ???



''' Wrrapers of pretrained models.
'''

import os

import rlcard
from rlcard.agents import CFRAgent
from rlcard.models.model import Model

# Root path of pretrianed models
ROOT_PATH = os.path.join(rlcard.__path__[0], 'models/pretrained')

class LeducHoldemNFSPModel(Model):
    ''' A pretrained model on Leduc Holdem with NFSP
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        import tensorflow as tf
        from rlcard.agents import NFSPAgent
        self.graph = tf.Graph()
        self.sess = tf.Session(graph=self.graph)

        env = rlcard.make('leduc-holdem')
        with self.graph.as_default():
            self.nfsp_agents = []
            for i in range(env.player_num):
                agent = NFSPAgent(self.sess,
                                  scope='nfsp' + str(i),
                                  action_num=env.action_num,
                                  state_shape=env.state_shape,
                                  hidden_layers_sizes=[128,128],
                                  q_mlp_layers=[128,128])
                self.nfsp_agents.append(agent)

        check_point_path = os.path.join(ROOT_PATH, 'leduc_holdem_nfsp')
        with self.sess.as_default():
            with self.graph.as_default():
                saver = tf.train.Saver()
                saver.restore(self.sess, tf.train.latest_checkpoint(check_point_path))
    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.nfsp_agents

class LeducHoldemNFSPPytorchModel(Model):
    ''' A pretrained PyTorch model on Leduc Holdem with NFSP
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        import torch
        from rlcard.agents import NFSPAgentPytorch
        env = rlcard.make('leduc-holdem')
        self.nfsp_agents = []
        for i in range(env.player_num):
            agent = NFSPAgentPytorch(scope='nfsp' + str(i),
                                     action_num=env.action_num,
                                     state_shape=env.state_shape,
                                     hidden_layers_sizes=[128,128],
                                     q_mlp_layers=[128,128],
                                     device=torch.device('cpu'))
            self.nfsp_agents.append(agent)

        check_point_path = os.path.join(ROOT_PATH, 'leduc_holdem_nfsp_pytorch/model.pth')
        checkpoint = torch.load(check_point_path)
        for agent in self.nfsp_agents:
            agent.load(checkpoint)

    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.nfsp_agents

class LeducHoldemCFRModel(Model):
    ''' A pretrained model on Leduc Holdem with CFR
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        env = rlcard.make('leduc-holdem')
        self.agent = CFRAgent(env, model_path=os.path.join(ROOT_PATH, 'leduc_holdem_cfr'))
        self.agent.load()
    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return [self.agent, self.agent]




import numpy as np


class RandomAgent(object):
    ''' A random agent. Random agents is for running toy examples on the card games
    '''

    def __init__(self, action_num):
        ''' Initilize the random agent

        Args:
            action_num (int): The size of the ouput action space
        '''
        self.use_raw = False
        self.action_num = action_num

    @staticmethod
    def step(state):
        ''' Predict the action given the curent state in gerenerating training data.

        Args:
            state (dict): An dictionary that represents the current state

        Returns:
            action (int): The action predicted (randomly chosen) by the random agent
        '''
        #return np.random.randint(0, self.action_num)
        return np.random.choice(state['legal_actions'])

    def eval_step(self, state):
        ''' Predict the action given the current state for evaluation.
            Since the random agents are not trained. This function is equivalent to step function

        Args:
            state (dict): An dictionary that represents the current state

        Returns:
            action (int): The action predicted (randomly chosen) by the random agent
            probs (list): The list of action probabilities
        '''
        probs = [0 for _ in range(self.action_num)]
        for i in state['legal_actions']:
            probs[i] = 1/len(state['legal_actions'])
        return self.step(state), probs
